/**
 * Licensed under the GNU LESSER GENERAL PUBLIC LICENSE, version 2.1, dated February 1999.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the latest version of the GNU Lesser General
 * Public License as published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program (LICENSE.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package org.jamwiki.model;

import java.sql.Timestamp;
import java.util.Date;
import java.util.Properties;
import org.jamwiki.Environment;
import org.jamwiki.WikiBase;
import org.jamwiki.utils.WikiLogger;

public class IntegritySettings implements IntegritySettingsI {
	private static final WikiLogger logger = WikiLogger.getLogger(IntegritySettingsHandler.class.getName());

	/* Number of different classification levels*/
	private int numberOfLevels = 5;
	/* Votes required to increase a userlevel */
	private int votesForUserInc = 10;
	/* How much over you own level are you allowed to vote for? 0 means you can only vote people to your own level*/
	private int votesMaxUserLevelDiff = 0;
	/* How long time until the vote ends */
	private int voteEndTimeDelta = 1000*60*60*24*14; //14 days
	private int voteExpireTimeDelta = 1000*60*60*24*21; //21 days
	
	private static final String RESOURCE_NAME = "/integrity.properties";
	
	/**
	 * Constructor
	 */
	IntegritySettings() {
		//Read properties from the resource file
		Properties p = Environment.loadProperties(RESOURCE_NAME);
		String temp = p.getProperty("numberOfLevels");
		if(temp != null) {
			try{
				numberOfLevels = Integer.parseInt(temp);
			}catch (Exception e) { }
		}
		temp = p.getProperty("votesForUserInc");
		if(temp != null) {
			try{
				votesForUserInc = Integer.parseInt(temp);
			}catch (Exception e) { }
		}
		temp = p.getProperty("votesMaxUserLevelDiff");
		if(temp != null) {
			try{
				votesMaxUserLevelDiff = Integer.parseInt(temp);
			}catch (Exception e) { }
		}
		temp = p.getProperty("voteEndTimeDelta");
		if(temp != null) {
			try{
				voteEndTimeDelta = Integer.parseInt(temp);
			}catch (Exception e) { }
		}
		temp = p.getProperty("voteExpireTimeDelta");
		if(temp != null) {
			try{
				voteExpireTimeDelta = Integer.parseInt(temp);
			}catch (Exception e) { }
		}
	}

    /**
     * Gives a description of the integrity settings implemented.
     * @return the string that can be put on the homepage as clean text
     */
    public String getDescription() {
        return	"The default Integrity settings are in effect!<br/>"+
				"There are currently 5 levels: 0-4 and a user can upgrade an article"+
				" up to his own level if he desires. <br/><br/>"+
				"Pages with classification level 0 are unprotected and can be edited by anonymous users.";
    }

    /**
     * Checks a vote for ended expired etc.
     * @param v the vote to be handled
     */
    public void handleVote(WikiVote v) {
        if(!v.isActive())
            return;
        if(v.ended(new Date())) {
			try{
				if(IntegrityTools.getVoteCount(v, -1, true)>=votesForUserInc && IntegrityTools.getVoteCount(v, -1, true)>IntegrityTools.getVoteCount(v, -1, false)) {
					//Vote passed
					IntegrityTools.changeUserLevel(v.getTarget(), v.getLevel());
					logger.info("Vote passed - User: "+v.getTarget().getUsername()+" moved to level "+v.getLevel());
				}
				//Now disable the vote
				v.setActive(false);
				IntegrityTools.updateCallvote(v);
			}catch (Exception e) {

			}
        }
    }

    /**
     * updates the logic. The actual logic is implementation dependent.
     */
    public void updateLogic() {

    }

    /**
     * Determens if a user is allowed to vote
     * @param wu The wikiuser that wants to vote
     * @param wv The vote the user wants to vote for
     * @return true if the user is allowed to vote.
     */
    public boolean userCanVote(WikiUser wu, WikiVote wv) {
        if(wv == null || wu==null)
            return false;
        if(wu.getClassification()>=wv.getLevel()-votesMaxUserLevelDiff)
            return true;
        return false;
    }

	/**
	 * Determens if a user is allowed to vote
	 * @param caller The wikiuser that wants to vote
	 * @param target The wikiuser that we want to vote about
	 * @return true if caller is allowed to vote
	 */
	public boolean userCanVote(WikiUser caller, WikiUser target) {
		if(caller == null || target==null)
			return false;
		if(caller.getActiveClassification()+votesMaxUserLevelDiff > target.getClassification())
			return true;
		return false;
	}

	/*
	 * We can never move a user down by vote in the basis model.
	 */
	public boolean userCanVoteDown(WikiUser caller, WikiUser target) {
		return false;
	}

    /**
     * Max level the topic may be moved up to by vote.
     * @param caller the user that wants to increase the level
     * @param id The topic that needs to get its level increased
     * @return the level it may be moved to.
     */
    public int maxLevelForTopic(WikiUser caller, Topic id) {
		if(caller==null)
			return 0;
        return caller.getClassification();
    }

	public int minLevelForTopic(WikiUser caller, Topic topic) {
		return topic.getClassification(); //We don't let the level go down
	}

    /**
     * The max level the user may be moved to by a single vote.
     * @param user the user in question
     * @return the level that the user may be moved to by vote.
     */
    public int maxLevelForUser(WikiUser user) {
		if(user.getClassification()<numberOfLevels)
			return user.getClassification()+1;
		return numberOfLevels;
    }

	/**
	 * Always returns the current level because we cannot move a user down by
	 * vote in the basis model
	 */
	public int minLevelForUser(WikiUser user) {
		return user.getClassification();
	}

	/**
	 * Fills end and expire dates in a vote based on starttime
	 * @param vote The vote to fill
	 */
	public void fillDates(WikiVote vote) {
		Date d = vote.getStart();
		vote.setEnd(new Timestamp(d.getTime()+voteEndTimeDelta));
		vote.setExpires(new Timestamp(d.getTime()+voteExpireTimeDelta));
	}
}
