/**
 * Licensed under the GNU LESSER GENERAL PUBLIC LICENSE, version 2.1, dated February 1999.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the latest version of the GNU Lesser General
 * Public License as published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program (LICENSE.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package org.jamwiki.model;

/**
 * This is the interface that you need to implement to make an alternative vote system
 * All methods must be overloaded. IntegritySettings.java can be used as a starting spot
 * @author poul
 */
public interface IntegritySettingsI {

    /**
     * Gives a description of the integrity settings implemented.
     * @return the string that can be put on the homepage as clean text
     */
    public String getDescription();

    /**
     * updates the logic. The actual logic is implementation dependent.
     */
    public void updateLogic();

    /**
     * Determens if a user is allowed to vote
     * @param wu The wikiuser that wants to vote
     * @param wv The vote the user wants to vote for
     * @return true if the user is allowed to vote.
     */
    public boolean userCanVote(WikiUser wu, WikiVote wv);

    /**
     * Max level the topic may be moved to
     * @param caller the user that wants to increase the level
     * @param topic The topic that needs to get its level increased
     * @return the level it may be moved to.
     */
    public int maxLevelForTopic(WikiUser caller, Topic topic);

	/**
	 * Min level the topic may be moved to
	 * @param caller the user that wants to decrease the level
	 * @param topic The topic that should have its level decreased
	 * @return the min level the topic may be moved to
	 */
	public int minLevelForTopic(WikiUser caller, Topic topic);

    /**
     * The max level the user may be moved to by a single vote.
     * @param user the user in question
     * @return the level that the user may be moved to by vote.
     */
    public int maxLevelForUser(WikiUser user);

	/**
	 * The min level the user may be moved to by vote (degraded)
	 * @param user the user in question
	 * @return the level the user may be moved to by vote.
	 */
	public int minLevelForUser(WikiUser user);

	/**
	 * Can the user's votes affect the user in question:
	 * @param caller The user that wants to call a vote
	 * @param target The user that will be voted for
	 * @return If the vote can be started
	 */
	public boolean userCanVote(WikiUser caller, WikiUser target);

	/**
	 * Can the user degrade another user by calling a vote?
	 * @param caller The user that wants to call a vote
	 * @param target The user that will be degraded should the vote pass
	 * @return If the vote can be started
	 */
	public boolean userCanVoteDown(WikiUser caller, WikiUser target);

	/**
	 * Fills end and expire dates in a vote based on starttime
	 * @param vote The vote to fill
	 */
	public void fillDates(WikiVote vote);

	/**
     * Checks a vote for ended expired etc.
     * @param v the vote to be handled
     */
    public void handleVote(WikiVote v);
}
