/**
 * Licensed under the GNU LESSER GENERAL PUBLIC LICENSE, version 2.1, dated February 1999.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the latest version of the GNU Lesser General
 * Public License as published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program (LICENSE.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.jamwiki.model;

import java.sql.Timestamp;
import org.jamwiki.WikiBase;
import java.util.Date;
import org.jamwiki.DataAccessException;
import org.jamwiki.WikiException;
import org.jamwiki.utils.WikiLogger;

/**
 *
 * @author poul
 */
public class IntegrityTools {
		private static final WikiLogger logger = WikiLogger.getLogger(IntegritySettingsHandler.class.getName());

       public IntegrityTools() {

       }

	   /**
		* Gets an ongoing vote for the current user
		* @param user The WikiUser that there is an onhoing vote about
		* @return The vote if it exists or null otherwise
		*/
	   static public WikiVote getVoteForUser(WikiUser user) {
		   WikiVote wv = null;
		   try {
				wv = WikiBase.getDataHandler().getWikiVoteUser(user);
		   } catch (Exception e) {
			   //Do nothing, just return null
		   }
		   return wv;
	   }

	   /**
		* Start a new vote.
		* @param caller The user that wants to start the vote
		* @param target The user that we want to vote about
		* @return true if the vote could be called, false otherwise
		*/
	   static public boolean callVote(WikiUser caller, WikiUser target) {
		   if(!IntegritySettingsHandler.getInstance().userCanVote(caller, target))
			   return false;
		   Date now = new Date();
		   WikiVote wv = new WikiVote(-1,new Timestamp(now.getTime()),target,true);
		   IntegritySettingsHandler.getInstance().fillDates(wv);
		   wv.setLevel(IntegritySettingsHandler.getInstance().maxLevelForUser(target));
		   wv.setCaller(caller);
		   try{
				WikiBase.getDataHandler().callVote(wv);
				logger.info("A vote about "+target.getUsername()+" was called by "+caller.getUsername()+" for moving to level "+wv.getLevel());
		   } catch (Exception e) {
			   //Do nothing... there are notrhing we can do except log the problem
			   logger.warning("Faild to start a vote about "+target.getUsername()+" by "+caller.getUsername()+" for moving to level "+wv.getLevel());
			   return false;
		   }
		   return true;
	   }

	   static public boolean callVoteDown(WikiUser caller, WikiUser target) {
		   if(!IntegritySettingsHandler.getInstance().userCanVoteDown(caller, target))
			   return false;
		   Date now = new Date();
		   WikiVote wv = new WikiVote(-1,new Timestamp(now.getTime()),target,true);
		   IntegritySettingsHandler.getInstance().fillDates(wv);
		   wv.setLevel(IntegritySettingsHandler.getInstance().minLevelForUser(target));
		   wv.setCaller(caller);
		   try{
				WikiBase.getDataHandler().callVote(wv);
				logger.info("A vote about "+target.getUsername()+" was called by "+caller.getUsername()+" for moving to level "+wv.getLevel());
		   } catch (Exception e) {
			   //Do nothing... there are notrhing we can do except log the problem
			   logger.warning("Faild to start a vote about "+target.getUsername()+" by "+caller.getUsername()+" for moving to level "+wv.getLevel());
			   return false;
		   }
		   return true;
	   }

	   /**
		* Cast a vote in a given election
		* @param voter The user that wants to vote
		* @param vote The vote that the user wants to vote for
		* @param yes true if the user wants to vote for the change, false otherwise
		*/
	   static public void castVote(WikiUser voter, WikiVote vote, boolean yes) {
           if(!IntegritySettingsHandler.getInstance().userCanVote(voter, vote))
               return;
		   try{
				WikiBase.getDataHandler().castVote(voter, vote, yes);
				logger.info(voter.getUsername()+" voted on the vote for target "+vote.getTarget().getUsername()+" voted yes: "+yes);
		   } catch (Exception e) {
			   //Do nothing... there are notrhing we can do
		   }
	   }

	   /**
		* Gets the count of votes for a given votenr
		* @param v The vote to count votes for
		* @param level The level of the users (votes) that we should count, -1 for all
		* @param yes true if we want to count 'yes' votes, false if we want to cound 'no' votes
		* @return Number of votes that match the criterias
		* @throws org.jamwiki.DataAccessException
		* @throws org.jamwiki.WikiException
		*/
	   static int getVoteCount(WikiVote v, int level, boolean yes) throws DataAccessException {
		   return WikiBase.getDataHandler().getVoteCount(v, level, yes);
	   }

	   /**
		* Updates a vote that has already started with some new values.
		* @param v The vote in question
		* @throws org.jamwiki.DataAccessException
		* @throws org.jamwiki.WikiException
		*/
	   static void updateCallvote(WikiVote v) throws DataAccessException {
			WikiBase.getDataHandler().updateCallvote(v);
	   }

	   /**
		* Changes a user's level
		* @param user The user for who we are going to change the level
		* @param newLevel The new level
		* @throws org.jamwiki.DataAccessException
		* @throws org.jamwiki.WikiException
		*/
	   static void changeUserLevel(WikiUser user, int newLevel)  throws DataAccessException {
		   WikiBase.getDataHandler().changeUserLevel(user, newLevel);
	   }

	   /**
		* Used to see when the user level was last changed. Possible null if no date (or user) could be found
		* @param userid The id of the user in question
		* @return Time the user level was last changed or null if not avaible
		*/
	   static Timestamp userLevelChanged(int userid) {
		   return WikiBase.getDataHandler().getUserLevelChanged(userid);
	   }

	   /**
		* Used to see when the topic level was last changed. Possible null if no date (or topic) could be found
		* @param topicid The id of the topic in quastion
		* @return Time the topic level was last changed or null if not avaible
		*/
	   static Timestamp topicLevelChanged(int topicid) {
		   return WikiBase.getDataHandler().getTopicLevelChanged(topicid);
	   }

	   /**
		 * Counts number of active user based on last active variable
		 * @param level Only count users on this level
		 * @param since Last active must be after this
		 * @return number of users matching criterias
		 * @throws org.jamwiki.WikiException If anything prevents us from returning an useable result
		 */
	   static int getActiveUserCountPerLevel(int level,Timestamp since) throws WikiException {
		   return WikiBase.getDataHandler().getActiveUserCountPerLevel(level, since);
	   }

	   /**
		* Counts number of changes done to topics by a given user
		* @param wikiuser_id The id of the user to check
		* @param since Time that the changes must be done after
		* @return number of changes by user in the given time
		* @throws org.jamwiki.WikiException If anything prevents us from getting an usable result
		*/
	   static int countTopicChangesPerUser(int wikiuser_id, Timestamp since) throws WikiException {
		   return WikiBase.getDataHandler().countTopicLevelChangesPerUser(wikiuser_id, since);
	   }
}
