/**
 * Licensed under the GNU LESSER GENERAL PUBLIC LICENSE, version 2.1, dated February 1999.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the latest version of the GNU Lesser General
 * Public License as published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program (LICENSE.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package org.jamwiki.servlets;

import java.util.Random;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;
import org.jamwiki.WikiMessage;
import org.jamwiki.utils.WikiLogger;
import org.jamwiki.utils.WikiUtil;
import org.jamwiki.WikiBase;
import org.jamwiki.model.Role;
import org.jamwiki.model.WikiUser;
import org.jamwiki.model.WikiVote;
import org.jamwiki.model.IntegritySettingsHandler;
import org.jamwiki.model.IntegrityTools;
import org.springframework.web.servlet.ModelAndView;

/**
 *
 * @author s032032
 */
public class UserClassificationServlet extends JAMWikiServlet {

    /** Logger for this class and subclasses. */
	private static final WikiLogger logger = WikiLogger.getLogger(SpecialPagesServlet.class.getName());
	/** The name of the JSP file used to render the servlet output. */
	protected static final String JSP_SPECIAL_PAGES = "user-classification.jsp";


    /**
	 * This method handles the request after its parent class receives control. It gets the topic's name and the
	 * virtual wiki name from the uri, loads the topic and returns a view to the end user.
	 *
	 * @param request - Standard HttpServletRequest object.
	 * @param response - Standard HttpServletResponse object.
	 * @return A <code>ModelAndView</code> object to be handled by the rest of the Spring framework.
	 */
	protected ModelAndView handleJAMWikiRequest(HttpServletRequest request, HttpServletResponse response, ModelAndView next, WikiPageInfo pageInfo) throws Exception {
		integrityPage(request, next, pageInfo);
		return next;
	}

    private void integrityPage(HttpServletRequest request, ModelAndView next, WikiPageInfo pageInfo) throws Exception {
		String userName = WikiUtil.getParameterFromRequest(request, "contributor", false);
		String callvote = WikiUtil.getParameterFromRequest(request, "callvote", false); //Do the user want to call a vote
		String callvote_down = WikiUtil.getParameterFromRequest(request, "callvote_down", false); //Do the user want to call a vote for degrading the user
		String newLevel = WikiUtil.getParameterFromRequest(request, "changeLevel", false); //Do the administrator wants to change the level
		String castvote = WikiUtil.getParameterFromRequest(request, "vote", false); //Voting on an existing vote
		String webkeyS = WikiUtil.getParameterFromRequest(request, "votekey", false); //Key to prevent crosssite attack
		Integer webkey = null;
		if(webkeyS != null)
			webkey = Integer.parseInt(webkeyS);
		//Get the user that the page is about:
		WikiUser user;
		if(userName != null)
			user = WikiBase.getDataHandler().lookupWikiUser(userName);
		else {
			//If no username is provided then default to admin (always userid = 1)
			user = WikiBase.getDataHandler().lookupWikiUser(1);
			userName = user.getUsername();
		}

		//Check if the user wants to call a vote.
		if(callvote!=null && user != null && callvote.equals("yes") && IntegritySettingsHandler.getInstance().userCanVote(ServletUtil.currentWikiUser(), user)
				&& IntegrityTools.callVote(ServletUtil.currentWikiUser(), user)
				&& request.getHeader("referer") != null && request.getHeader("referer").startsWith(request.getContextPath())) {
			pageInfo.setPageTitle(new WikiMessage("specialpages.caption.callvote"));
			next.addObject("votecalled",1);
		} else if(callvote_down!=null && user != null && callvote_down.equals("yes") &&
				IntegritySettingsHandler.getInstance().userCanVoteDown(ServletUtil.currentWikiUser(), user) &&
				IntegrityTools.callVoteDown(ServletUtil.currentWikiUser(), user)){
			pageInfo.setPageTitle(new WikiMessage("specialpages.caption.callvote"));
			next.addObject("votecalled",1);
		} else { //The user does not want to call a vote... yet
			pageInfo.setPageTitle(new WikiMessage("tab.common.userclass"," - "+userName));
			next.addObject("votecalled",0);
		}
		//Try to get en existing vote if such exist
		WikiVote wv = null;
		if(user!=null)
			wv = IntegrityTools.getVoteForUser(user);

		//If the administrator has provided the 'changeLevel' parameter then change the level
		if(newLevel!=null && user!=null && request.getHeader("referer") == null && ServletUtil.currentUserDetails().hasRole(Role.ROLE_ADMIN)) {
			int level = Integer.parseInt(newLevel);
			WikiBase.getDataHandler().changeUserLevel(user,level);
		}
		HttpSession session = request.getSession(false);
		Integer key;
		if(session.getAttribute("votekey")==null)
		{
			key = (new Random()).nextInt();
			session.setAttribute("votekey", key);
		} else {
			key = (Integer)session.getAttribute("votekey");
		}
		next.addObject("votekey", key);
		if(castvote!=null && wv!=null) {
			if( webkey != null &&
					webkey.equals(key) ) {
				if(castvote.equalsIgnoreCase("yes")) {
					IntegrityTools.castVote(ServletUtil.currentWikiUser(), wv, true);
					next.addObject("votecasted",1);
				} else {
					if(castvote.equalsIgnoreCase("no")) {
						IntegrityTools.castVote(ServletUtil.currentWikiUser(), wv, false);
						next.addObject("votecasted",-1);
					} else {
						next.addObject("votecasted",0);
					}
				}
			} else {
				//Write error to log file. If there is a lot of these entries in the log then the wiki might be under attack
				logger.warning("Failed to cast a vote due to key failure! Session key is: "+key+" provided key is: "+webkey+", voting user: "+
						ServletUtil.currentWikiUser().getUsername()+", target user: "+userName);
			}
		}
		if(ServletUtil.currentUserDetails().hasRole(Role.ROLE_ANONYMOUS) || ServletUtil.currentWikiUser()==null) {
			//Anonymous can never vote
			next.addObject("canCallVote", 0);
			next.addObject("canvote", 0);
			next.addObject("canCallVoteDown", 0);
		}
		else {
			//Ask the IntegritySettings if the user can vote and add objects that tells jsp that he can
			next.addObject("canvote", IntegritySettingsHandler.getInstance().userCanVote(ServletUtil.currentWikiUser(), user)?1:0);
			next.addObject("canCallVote", IntegritySettingsHandler.getInstance().userCanVote(ServletUtil.currentWikiUser(), user)?1:0);
			next.addObject("canCallVoteDown", IntegritySettingsHandler.getInstance().userCanVoteDown(ServletUtil.currentWikiUser(), user)?1:0);
		}
		if(wv==null) { //There is no vote in progress
			next.addObject("voteInProgress", 0);
		} else { //There is a vote in progress
			IntegritySettingsHandler.getInstance().handleVote(wv);
			next.addObject("voteInProgress", 1);
			next.addObject("datestarted", wv.getStart());
			next.addObject("dateends", wv.getEnd());
			next.addObject("votesFor", WikiBase.getDataHandler().getVoteCount(wv, -1, true)); //Count all votes for
			next.addObject("votesAgainst",WikiBase.getDataHandler().getVoteCount(wv, -1, false)); //Count all votes against
			next.addObject("newlevel",wv.getLevel());
		}
		//General information
		if(user!=null)
			next.addObject("currentlevel",user.getClassification());
		else
			next.addObject("currentlevel",0);
		next.addObject("username",userName);
		//Special page:
		pageInfo.setContentJsp(JSP_SPECIAL_PAGES);
		pageInfo.setSpecial(true);
	}

}
